import java.util.Comparator;
import java.lang.Math;
/**
 * Clase para trabajar con arboles binarios de busqueda balanceados, llamados AVL
 * @author Amparo Lopez Gaona
 * @author Yazmin Gonzalez Bayardo
 * @version Septiembre 2015
 */   
public class ArbolAvl {
	
    private NodoAvl raiz;            // Nodo raiz del arbol
    private Comparator<Object> cmp;  // Comparador 
    private int nNodos;              // Numero de elementos en el arbol
	
    /**
     * Constructor a partir de un comparador.
     * @param c - comparator para establecer relacion de orden entre nodos.
     */
    public ArbolAvl(Comparator<Object> c) {
		raiz = null;
		cmp = c;
		nNodos = 0;
    }

    /**
     * Metodo para dejar vacio un arbol.
     */
    public void vaciar() {
		raiz = null;
		nNodos = 0;
    }

    /**
     * Metodo para determinar si un arbol esta vacio.
     * @return boolean - true si el arbol esta vacio y false en otro caso.
     */
    public boolean estaVacio() {
		return raiz == null;
    }

    /**
     * Metodo para conocer el tamano de un arbol.
     * @return int - cantidad de elementos en el arbol.
     */
    public int tamanio() {
		return nNodos;
    }
	
    /**
     * Metodo para insertar un nodo en el arbol, ignorando los duplicados y 
     * balanceando si es necesario.
     * @param dato - el elemento a insertar.
     */
    public void agregar(Object dato) {
		raiz = agregar(dato, raiz);
    }
	
    /**
     * Metodo para encontrar un elemento en el arbol.
     * @param dato - el dato a buscar.
     * @return boolean - true si el elemento se encontro o false si no esta.
     */
    public boolean contiene(Object dato) {
		return encontrar(dato, raiz) != null;
    }
    
    /**
     * Metodo interno, auxiliar, para agregar en un arbol.
     * @param dato - elemento a agregar.
     * @param n - nodo raiz del arbol.
     * @return NodoAvl - la nueva raiz.
     */
    private NodoAvl agregar(Object dato, NodoAvl n) {
		
		if(n == null) {
			n = new NodoAvl(dato);
			nNodos ++;
		}
		else if(cmp.compare(dato, n.valor) < 0) {
			n.izquierda = agregar(dato, n.izquierda);
			if(altura(n.izquierda) - altura(n.derecha) == 2) {
				if(cmp.compare(dato, n.izquierda.valor) < 0) // Rotar sobre izquierda
					n = rotarIzq(n);
				else { // Doble rotacion
					n.izquierda = rotarDer(n.izquierda);
					n = rotarIzq(n);
				}
			}
		} else if(cmp.compare(dato, n.valor) > 0) {
			n.derecha = agregar(dato, n.derecha);
			if(altura(n.derecha) - altura(n.izquierda) == 2){
				if(cmp.compare(dato, n.derecha.valor) > 0) // Rotar sobre derecha
				    n = rotarDer(n);
				else { // Doble rotacion
				    n.derecha = rotarIzq(n.derecha);
				    n = rotarDer(n);
				}
			}
		} else ;  // Encontro un duplicado y no hace nada.
		
		n.altura = Math.max(altura(n.izquierda), altura(n.derecha)) + 1;
		return n;
    }
	
    /**
     * Metodo privado para conocer la altura de un nodo.
     * @param n - Nodo del que se desea conocer la altura.
     * @return int - altura del nodo.
     */
    private int altura (NodoAvl n) {
		return (n == null) ? -1 : n.altura;
    }

    /**
     * Metodo para rotar sobre subarbol izquierdo.
     * @param n - nodo raiz del subarbol que se va a rotar.
     * @return NodoAvl - Nodo raiz del subarbol despues de la rotacion.
     */
    private NodoAvl rotarIzq(NodoAvl n) {
		NodoAvl nraiz = n.izquierda;
		n.izquierda = nraiz.derecha;
		nraiz.derecha = n;
		n.altura = Math.max(altura(n.izquierda), altura(n.derecha)) + 1;
		nraiz.altura = Math.max(altura(nraiz.izquierda), n.altura) + 1;
		return nraiz;
    }

    /**
     * Metodo para rotar sobre subarbol derecho.
     * @param n - nodo raiz del subarbol que se va a rotar.
     * @return NodoAvl - Nodo raiz del subarbol despues de la rotacion.
     */
    private NodoAvl rotarDer(NodoAvl n) {
		NodoAvl nraiz = n.derecha;
		n.derecha = nraiz.izquierda;
		nraiz.izquierda = n;
		n.altura = Math.max(altura(n.izquierda), altura(n.derecha)) + 1;
		nraiz.altura = Math.max(altura(nraiz.derecha), n.altura) + 1;
		return nraiz;
    }
	
    /**
     * Metodo interno para encontrar el menor elemento en un subarbol.
     * @param n - nodo raiz del arbol.
     * @return NodoAVl - el nodo que contiene el elemento menor.
     */
    private NodoAvl encuentraMin(NodoAvl n) {
        if(n == null)
            return null;
        else if(n.izquierda == null)
            return n;
        return encuentraMin(n.izquierda);
    }
	
    /**
     * Metodo interno para encontrar el mayor elemento en un subarbol.
     * @param n - raiz del arbol.
     * @return NodoAvl - el nodo que contiene el elemento mayor.
     */
    private NodoAvl encuentraMax(NodoAvl n) {
        if(n != null)
            while(n.derecha != null)
                n = n.derecha;
        return n;
    }

    /**
     * Metodo interno para encontrar un elemento en un subarbol.
     * @param dato - elemento buscado.
     * @param n - raiz del arbol.
     * @return NodoAVL - que contiene el elemento encontrado o null si no lo encontro.
      */
    private NodoAvl encontrar(Object dato, NodoAvl n) {
		while(n != null){
		    if(cmp.compare(dato, n.valor) < 0)
			n = n.izquierda;
		    else if(cmp.compare(dato, n.valor) > 0)
			n = n.derecha;
		    else
			return n;    // Lo encontro
		}
		
		return null;   // No lo encontro.
    }
	
    /**
     * Metodo para eliminar un elemento del arbol, en caso de no encontrarlo no hace nada.
     * @param dato - el dato a eliminar.
     */
    public void eliminar(Object dato) {
		raiz = eliminar(dato, raiz);
    }
	
	/**
     * Metodo interno para eliminar un elemento de un subarbol.
     * @param elemento - el dato a eliminar.
     * @param nodo - raiz del arbol.
     * @return NodoAvl - la nueva raiz.
     */
    private NodoAvl eliminar(Object dato, NodoAvl n) {
        if(n == null)
           return n;   // No se encuentra el elemento; no hace nada
		
        if(cmp.compare(dato, n.valor) < 0)
           n.izquierda = eliminar(dato, n.izquierda);
        else if(cmp.compare(dato, n.valor) > 0)
           n.derecha = eliminar(dato, n.derecha);
        else {
			if(n.izquierda != null && n.derecha != null) { // Nodo con dos hijos
				n.valor = encuentraMin(n.derecha).valor; // Obtiene sucesor (elemento mas pequeno del subarbol derecho)
				n.derecha = eliminar(n.valor, n.derecha);
			}
			else { // Nodo con un solo hijo o sin hijos
				n = (n.izquierda != null) ? n.izquierda : n.derecha;
				nNodos--;
			}
        } 
		
		if(n==null)
			return n;
		
		n.altura = Math.max(altura(n.izquierda), altura(n.derecha)) + 1;
		
		if(altura(n.izquierda) - altura(n.derecha) == 2) { //left heavy
			if((altura(n.izquierda.izquierda) - altura(n.izquierda.derecha)) >= 0) { // Rotar sobre izquierda
				n = rotarIzq(n);
				return n;
			}
			
			if((altura(n.izquierda.izquierda) - altura(n.izquierda.derecha)) < 0) { // Doble rotacion
				n.izquierda = rotarDer(n.izquierda);
				n = rotarIzq(n);
				return n;
			}		
		}
		
		if(altura(n.derecha) - altura(n.izquierda) == 2) { //right heavy
			if((altura(n.derecha.izquierda) - altura(n.derecha.derecha)) <= 0 ) { // Rotar sobre derecha
				n = rotarDer(n);
				return n;
			}
			
			if((altura(n.derecha.izquierda) - altura(n.derecha.derecha)) > 0 ) { // Doble rotacion
			    n.derecha = rotarIzq(n.derecha);
			    n = rotarDer(n);
				return n;
			}
		}
		
        return n;
    }
	
    /**
     * Metodo para imprimir el contenido del arbol en inOrden.
     */
    public void imprimirArbol() {
        if(estaVacio())
            System.out.println("El arbol esta vacio.");
        else
            imprimirArbol(raiz);
		System.out.println();
    }

    /**
     * Metodo auxiliar para imprimir el arbol en inOrden.
     * @param nodo - la raiz del arbol.
     */
    private void imprimirArbol(NodoAvl nodo) {
		if(nodo != null) {
			imprimirArbol(nodo.izquierda);
			System.out.println(nodo.valor +":"+nodo.altura+"\t");
            imprimirArbol(nodo.derecha);
        }
    }
}















