import java.util.Comparator;
import java.util.Iterator;
import java.util.NoSuchElementException;
/**
 * Clase para crear y trabajar colas de prioridad utilizando heaps.
 * @author Amparo Lopez Gaona
 * @author Yazmin Gonzalez Bayardo
 * @version Noviembre 2015
 */
public class Heap implements InterfazColaConPrioridad{
	
    private Object [] elementos;   // Arreglo en donde se guarda la cola
    private int nElementos ;       // Cantidad de elementos almacenados
    private Comparator<Object> cmp;        // Comparador para los elementos
	
	/**
     * Constructor que recibe un comparador. Asigna el tamano del heap para 100 elementos.
     * @param c - comparador para la prioridad de los elementos.
     */
	public Heap (Comparator<Object> c) { 
		this(c, 100);
	}
	
  	/**
   	 * Constructor que recibe un comparador y el tamano maximo del heap.
   	 * @param c - comparador para la prioridad de los elementos.
   	 * @param tam - tamano máximo del heap.
   	 */
    public Heap(Comparator<Object> c, int tam) {
		cmp = c;
		elementos = new Object[(tam > 0) ? tam : 100];
		nElementos = 0;
    }
	
	/**
   	 * Constructor que crea una cola de prioridad a partir de un arreglo.
   	 * @param c - comparador para la prioridad de los elementos.
   	 * @param datos - arreglo de objetos que formarán el heap.
   	 */
    public Heap(Comparator<Object> c, Object [] datos) {
		cmp = c;
		nElementos = datos.length;
		
		elementos = new Object[nElementos*2];
		for (int i = 0; i < nElementos; i++)
			elementos[i] = datos[i];
		
		crearHeap(elementos);
    }

  	/**
   	 * Constructor de copia.
   	 * @param hep - heap del cual se creara una copia.
   	 */
  	public Heap(Heap hep) {
		nElementos = hep.nElementos;
	  	elementos = new Object[nElementos];
	  	cmp = hep.cmp;
	  	for (int i = 0; i < nElementos; i++)
			elementos[i] = hep.elementos[i];
  	}

  	/**
   	 * Metodo para determinar si la cola de prioridad esta vacia.
   	 * @return boolean - devuelve true si la cola está vacía y false en otro caso.
   	 */
    public boolean estaVacia() { 
		return nElementos == 0;
    }

  	/**
   	 *  Metodo para eliminar todos los elementos de la cola de prioridad.  
   	 */
    public void vaciar() {
		nElementos = 0;
    }

  	/** 
   	 * Metodo para conocer la cantidad de elementos que tiene la cola de prioridad.
   	 * @return int - número de elementos en la cola de prioridad.
   	 */
    public int tamanio() { 
		return nElementos;
    }

  	/**
   	 *  Metodo para obtener el primer elemento de la cola de prioridad.
   	 * @return Object - el primer elemento en la cola de prioridad.
   	 */
    public Object obtenerPrimero() {  
		return elementos[0];
    }

  	/**
   	 * Metodo para agregar un nuevo elemento en la cola de prioridad.
   	 * @param valor - objeto que se va a insertar.
   	 */
    public void agregar(Object valor) { 
		int posicion = nElementos; // Va a agregar el dato al final
        int posPadre = (nElementos-1)/2;
		Object valorPadre = null;  

		if(posicion == elementos.length)
			crecerArreglo();
		
		elementos[posicion] = valor;
		if (posPadre >= 0)
			valorPadre = elementos[posPadre];
		
		while((posicion > 0) && (cmp.compare(valor, valorPadre) < 0)) { //Min-Heap (cola de prioridad minima)
			elementos[posicion] = valorPadre;
	    	posicion = posPadre;
	    	posPadre = (posicion-1)/2;
	    	if(posPadre >= 0)
				valorPadre = elementos[posPadre];
		}
		elementos[posicion] = valor;
		nElementos++;
    }
	
    /**
     * Metodo privado para aumentar la capacidad del arreglo.
     */
    private void crecerArreglo() {
		System.out.println("Voy a aumentar el tamano del arreglo");
		Object [] tmp = new Object[elementos.length+10];
		for (int i = 0; i < elementos.length; i++)
	    	tmp[i] = elementos[i];
		elementos = tmp;
    }
	
	/**
	 * Metodo para eliminar el elemento de menor prioridad.
   	 */
    public void eliminar() {
		if(tamanio() > 0) { // Hay al menos un elemento por eliminar
			int ultimaPos = tamanio() - 1;
		
			if(ultimaPos != 0)  // Elimina el primer elemento
				elementos[0] = elementos[ultimaPos];

			nElementos--;  // Elimina el ultimo elemento
			ajustarHeap(0, nElementos);
		}
    }

	/**
	 * Metodo privado para ajustar un heap (hacia abajo a partir de la posicion indicada).
	 * @param pos - posicion del nodo a partir del cual va a verificar que se tenga un heap.
	 * @param tamano - tamanio del arreglo.
	 */
    private void ajustarHeap(int pos, int tamano) {
		Object valor = elementos[pos];
		while (pos < tamano) {
			int posHijo = pos*2 + 1;
	    	if (posHijo < tamano) {
				if (((posHijo + 1) < tamano) && 
					cmp.compare(elementos[posHijo+1], elementos[posHijo]) <0) // Se toma el menor de los hijos
						posHijo++;
				if (cmp.compare(valor, elementos[posHijo]) < 0) {  // Min-heap (cola de prioridad minima)
					elementos[pos] = valor;
		    		return;
				}
				else {
					elementos[pos] = elementos[posHijo];
					pos = posHijo;
				}
			}
			else {
				elementos[pos] = valor;
				return;
	    	}
		}
    }
	
	/**
	 * Metodo privado para crear un heap a partir de un arreglo.
	 * @param a - arreglo con los objetos que se almacenaran en el heap.
	 */
	private void crearHeap(Object[] a) {
		int max = nElementos;
		for(int i = max/2; i >= 0; i--)
			ajustarHeap(i,max);
	}

	
	/**
	 * Metodo para imprimir el heap como un arbol.
	 */
	public void imprimirHeap() {
		for(int i=0; i<nElementos; i++)
			System.out.print(elementos[i]+" ");
		
		System.out.println();
	}
	
	/**
	 * Metodo para cambiar el elemento en la posicion p con un nuevo valor. No hace nada si la posicion es invalida.
	 * @param p - posicion del elemento.
	 * @param nuevoValor - nuevo elemento.
	 */
	public void cambiar(int p, Object nuevoValor) {
		if(!estaVacia() && p>=0 && p<=tamanio()-1) {
			
			elementos[p] = nuevoValor; //Sustituye por el nuevo valor
			
			ajustarArriba(p);
			ajustarHeap(p,nElementos);
		}	
	}
	
	/**
	 * Metodo privado para ajustar un heap (hacia arriba a partir de la posicion indicada).
	 * @param p - posicion del nodo a partir del cual va a verificar que se tenga un heap.
	 */
	private void ajustarArriba(int p) {
		int posPadre = (p-1)/2;
		while(p>0 && posPadre>=0 && (cmp.compare(elementos[p],elementos[posPadre]) < 0)) {
				intercambiar(p, posPadre);
				p = posPadre;
				posPadre = (p-1)/2;
		}
	}
	
	/**
	 * Metodo privado para intercambiar el valor de los nodos en las posiciones indicadas como parametros.
	 * @param pos - posicion del nodo hijo.
	 * @param posPadre - posicion del nodo padre.
	 */
	private void intercambiar(int pos, int posPadre) {
		Object tmp = elementos[pos];
		elementos[pos] = elementos[posPadre];
		elementos[posPadre] = tmp;
	}
	
	
	/**
	 * Metodo para eliminar el elemento en la posicion p del arreglo. No hace nada si la posicion es invalida.
	 * @param p - posicion del elemento por eliminar.
	 */
	public void borrar(int p) {
		
		int ultimaPos = tamanio()-1;
		if(!estaVacia() && p>=0 && p<=ultimaPos) { // Checar que sea una posicion valida
			
			if(p != ultimaPos){
				elementos[p] = elementos[ultimaPos]; // Elemento por eliminar se sustituye con el ultimo elemento
				nElementos--;
				
				int posPadre = (p-1)/2;
				if((p > 0) && cmp.compare(elementos[p],elementos[posPadre]) < 0) // Ajustar hacia arriba si se sustituyo con un elemento mayor que el del padre
					ajustarArriba(p);
				else
					ajustarHeap(p, nElementos); //Ajustar hacia abajo en otro caso
			}
			else
				nElementos--; // Solo elimina el ultimo elemento, nada por hacer
		}
	}
	
	/**
	 * Metodo que devuelve un iterador sobre la cola de prioridad en orden ascendente.
	 * @return Iterator - iterador sobre la cola de prioridad.
	 */
	public Iterator iterador() {
		return new HeapIterador(this);
	}
	
	/**
	 * Clase interna que implementa el iterador sobre la cola de prioridad en orden ascendente.
	 */
	private class HeapIterador implements java.util.Iterator {
		
		private Heap copia;
		private Object primero;
		
		public HeapIterador(Heap heap) {
			copia = new Heap(heap);
		}
		
		public boolean hasNext()  {
			return !copia.estaVacia();
		}
		
		public Object next(){
			if(hasNext()){
				primero = copia.obtenerPrimero();
				copia.eliminar();
				return primero;
			}
			throw new NoSuchElementException("No hay elementos en el heap.");
		}
		
		// Metodo no soportado
		public void remove() {
			throw new UnsupportedOperationException();
		}
	}
}