import java.util.Iterator;
import java.util.Scanner;
import java.io.File;
import java.io.FileNotFoundException;
/**
 * Clase para simular intercambio de acciones de una compania, ya sea compra o venta de acciones haciendo uso de colas de prioridad.
 * @author Yazmin Gonzalez Bayardo
 * @version Noviembre 2015
 */
public class MercadoAcciones {
	
	private Heap ordenesVenta; // Cola de prioridad minima para ordenes de venta de acciones.
 	private Heap ordenesCompra; // Cola de prioridad maxima para ordenes de compra de acciones.
	private StringBuilder registro; // Almacena registro de las transacciones realizadas y las ordenes pendientes.
	
	/**
	 * Constructor para la simulacion de compra y venta de acciones.
	 */
	public MercadoAcciones() {
		ordenesVenta = new Heap(new ComparaOrdenesVenta());
	 	ordenesCompra = new Heap(new ComparaOrdenesCompra());
	}
	
	/**
	 * Metodo que lee de un archivo las ordenes de compra y venta de acciones. Procesa cada orden y lleva un registro de las transacciones realizadas.
	 * Al finalizar la lectura del archivo muestra en pantalla el registro de las transacciones, asi como las ordenes de compra y venta que no pudieron completarse.
	 * @param nombreArchivo - archivo de texto que contiene las ordenes de compra y venta. 
	 */
	public void simular(String nombreArchivo) {
		
		try{
			ordenesVenta.vaciar();
			ordenesCompra.vaciar();
			registro = new StringBuilder();
			
			Scanner sc = new Scanner(new File(nombreArchivo));
			
			ComparaOrdenesVenta cmpGeneral = new ComparaOrdenesVenta(); //Comparador para calcular diferencia de precio entre orden de venta y orden de compra
			
			while(sc.hasNextLine()) {
				String linea = sc.nextLine();
				
				String[] datosOrden = linea.split(",");
				
				if(datosOrden.length < 4)
					throw new RuntimeException("Orden con datos incompletos.");
				
				String usuario = datosOrden[0];
				String tipoOrden = datosOrden[1];
				int numAcciones = Integer.parseInt(datosOrden[2]);
				int precio = Integer.parseInt(datosOrden[3]);
				
				Orden orden = new Orden(usuario, tipoOrden, precio, numAcciones);
				
				if(tipoOrden.equals("compra")) //Compra de acciones
					ordenesCompra.agregar(orden);
				else {
					if(tipoOrden.equals("vende")) //Venta de acciones
						ordenesVenta.agregar(orden);
					else 
						throw new RuntimeException("Tipo de orden invalido en el archivo.");
				}
					
				if( ordenesVenta.tamanio() == 0 || ordenesCompra.tamanio() == 0 )continue; //Continuar leyendo el archivo pues no hay ordenes por comparar.
				
				Orden venta = (Orden)ordenesVenta.obtenerPrimero(); // Obtiene la orden de venta con menor precio por accion.
				Orden compra = (Orden)ordenesCompra.obtenerPrimero(); // Obtiene la orden de compra con mayor precio por accion.
				
				if(cmpGeneral.compare(venta,compra) <= 0 ) // Procesar ordenes si el precio de la orden de venta es menor o igual que el de la orden de compra.
					procesarOrden(venta,compra);
			}
			
			mostrar(); // Mostrar transacciones realizadas y ordenes pendientes.
			
		}catch(FileNotFoundException e){
			System.out.println(e + " Archivo no encontrado.");
		}catch(Exception e){
			System.out.println(e);
		}
	}
	
	/**
	 * Metodo que procesa las ordenes para completarla transaccion de compra y venta de acciones.
	 * @param venta - orden de venta de acciones.
	 * @param compra - orden de compra de acciones.
	 */
	private void procesarOrden(Orden venta, Orden compra) {
			
			int numAccVenta = venta.obtenerNumAcciones();
			int numAccCompra = compra.obtenerNumAcciones();
			
			if(numAccVenta < numAccCompra) { // Elimina la orden de venta y actualizar numero de acciones restantes por comprar.
				
				ordenesVenta.eliminar();
				compra.asignarNumAcciones(numAccCompra-numAccVenta);
				
				// Registra la transaccion realizada.
				registro.append(compra.obtenerUsuario() + " compra a " + venta.obtenerUsuario() +
					 " acciones:" + venta.obtenerNumAcciones() + " precio: $" + compra.obtenerPrecio() + "\n");
				
				// Trata de completar la compra del resto de las acciones.
				if( ordenesVenta.tamanio() > 0){
					ComparaOrdenesVenta cmp = new ComparaOrdenesVenta(); //Comparador para calcular diferencia de precio entre la orden de venta y la orden de compra
					Orden v = (Orden)ordenesVenta.obtenerPrimero();
					Orden c = (Orden)ordenesCompra.obtenerPrimero();
				
					if(cmp.compare(v,c) <= 0 )
						procesarOrden(v,c);
				} 
				
				
			}else{
				if(numAccVenta > numAccCompra) { // Elimina la orden de compra y actualiza el numero de acciones restantes por vender.
					
					ordenesCompra.eliminar();
					venta.asignarNumAcciones(numAccVenta-numAccCompra);
				}
				else // Eliminar ambas ordenes de compra y venta.
				{
					ordenesVenta.eliminar();
					ordenesCompra.eliminar();
				}
				
				// Registra la transaccion realizada.
				registro.append(compra.obtenerUsuario() + " compra a " + venta.obtenerUsuario() +
					 " acciones:" + compra.obtenerNumAcciones() + " precio: $" + compra.obtenerPrecio() + "\n");
			}
	}
	
	/**
	 * Metodo privado para mostrar en pantalla las transacciones realizadas y las ordenes de compra y venta que no pudieron completarse.
	 */
	private void mostrar() {
		
		registro.append("\n----Ordenes de Venta pendientes: \n");
	
		if(ordenesVenta.estaVacia()){
			registro.append("\n");
		}else{
			Orden o;
			Iterator itV = ordenesVenta.iterador();
			while(itV.hasNext()){
				o = (Orden)itV.next();
				registro.append(o.obtenerUsuario() + " acciones:" + o.obtenerNumAcciones() + " precio: $" + o.obtenerPrecio() +" \n");
			}
		}

		registro.append("\n----Ordenes de Compra pendientes: \n");
		if(ordenesCompra.estaVacia()){
			registro.append("\n");
		}else{
			Orden o;
			Iterator itC = ordenesCompra.iterador();
			while(itC.hasNext()){
				o = (Orden)itC.next();
				registro.append(o.obtenerUsuario() + " acciones:" + o.obtenerNumAcciones() + " precio: $" + o.obtenerPrecio() +" \n");
			}
		}
		registro.append("\n");
		System.out.print(registro.toString());
	}
}