/**
 * Clase para crear tablas de dispersión usando el metodo lineal para resolver colisiones y considera el metodo eliminar.
 * @author Yazmin Gonzalez Bayardo
 * @version Diciembre 2015
 */
public class TablaDeDispersionA2 implements InterfazTablaDispersion {
	private Object[] elementos;
  	private int nElementos;
	private final Object borrado = "BORRADO";
	
	/**
	 * Crea una tabla de dispersion del tamanio especificado.
	 * @param tam -- cantidad inicial de elementos permitida.
	 */
	public TablaDeDispersionA2 (int tam) {
		elementos = new Object[sgtePrimo(tam)];
		nElementos = 0;
	}
	
	/**
	 * Constructor por omision. Crea una tabla de dispersion para 101 elementos.
	 */
	public TablaDeDispersionA2 () {
		this(100);
	}
	
	/**
	 * Verifica si la tabla esta vacia.
	 * @return boolean -- true si la tabla esta vacia y false en otro caso.
	 */
	public boolean estaVacia () {
		return nElementos == 0;
	}
	
	/**
	 * Determina la cantidad de elementos en la tabla.
	 * @return int -- cantidad de elementos en la tabla.
	 */
	public int tamanio () {
		return nElementos;
	}
	
	/**
	 * Elimina todos los elementos de la tabla.
	 */
	public void vaciar () {
		for (int i = 0; i < elementos.length; i++)
			elementos[i] = null;
		nElementos = 0;
	}
	
	/**
	 * Devuelve la capacidad de la tabla.
	 * @return int -- capacidad de la tabla para almacenar elementos.
	 */
	public int capacidad () {
		return elementos.length;
	}
	
	/**
	 * Metodo privado para ajustar la capacidad de la tabla segun el parametro indicado. 
	 * Los elementos actuales se reasignan en la tabla de dispersion ajustada.
	 * @param capacidad -- nueva capacidad de la tabla.
	 */
	private void ajustarTabla(int capacidad) {
		
		TablaDeDispersionA2 temp = new TablaDeDispersionA2(capacidad);
		
		for(int i=0; i<elementos.length; i++)
		{
			Object elem = elementos[i];
			if(elem!=null && !elem.equals(borrado))
				temp.agregar(elementos[i]);
		}
		this.elementos = temp.elementos;
	}
	
	/**
	 * Metodo para mostrar el contenido completo de la tabla.
	 */
    public void imprimir() {
		for(int i=0; i<elementos.length; i++)
  	    	System.out.print(i+":"+elementos[i]+"  ");
  		System.out.println();
    }
	
	/**
	 * Agrega un elemento a la tabla.
	 * @param val -- objeto que se insertará en la tabla.
	 */
	public void agregar (Object val) {
		
		if(nElementos >= elementos.length/2) // Checa si la mitad de la tabla esta ocupada; si es asi entonces duplica la capacidad de la tabla.
			ajustarTabla(2*elementos.length);
		
		int indice = Math.abs(val.hashCode()) % elementos.length;
		
		while (elementos[indice] != null && !elementos[indice].equals(borrado)) //Resolución de colisiones
				if (++indice >= elementos.length) indice = 0;
		
		elementos[indice] = val;
    	nElementos++;
	}
	
    /**
     * Metodo para eliminar el valor asociado a una llave.
     * @param val - llave del elemento que será eliminado de la tabla.
     */
    public void eliminar (Object val) {
    	int indiceIni = Math.abs(val.hashCode()) % elementos.length;
    	int indice = indiceIni;
		
		while (elementos[indice] != null) {
			if(!elementos[indice].equals(borrado)){
				if (val.equals(elementos[indice])) {
					elementos[indice] = borrado;
				  	nElementos--;
					if(nElementos>0 && nElementos<=elementos.length/8)
						ajustarTabla(elementos.length/2); //Reduce la capacidad de la tabla si solo un octavo la tabla o menos esta siendo usada.
				  	break;
				}
			}
			if (++indice >= elementos.length) indice = 0;
			if (indiceIni == indice) break;
    	}
    }
	
	/**
   	 * Determina si una tabla contiene un valor particular.
   	 * @param val -- elemento que se busca.
   	 * @return boolean -- true si la tabla contiene un valor particular y false en otro caso.
   	 */
	public boolean contiene (Object val) {
		int indiceIni = Math.abs(val.hashCode()) % elementos.length;
		int indice = indiceIni;
		
		while (elementos[indice] != null) {
			if(!elementos[indice].equals(borrado)){
				if (val.equals(elementos[indice]))
					return true;
			}
			if (++indice >= elementos.length) indice = 0;
			if (indiceIni == indice) break;
		}
		return false;
	}
	
	/**
	 * Metodo para obtener el valor asociado a una llave.
	 * @param val -- llave del elemento que se busca.
	 * @return Object -- objeto asociado a la llave o null si no lo encuentra.
	 */
	public Object obtener (Object val) {
    	int indiceIni = Math.abs(val.hashCode()) % elementos.length;
    	int indice = indiceIni;
		
		while (elementos[indice] != null) {
			if(!elementos[indice].equals(borrado)){
				if (val.equals(elementos[indice]))
		  		  return elementos[indice];
			}
			if (++indice >= elementos.length) indice = 0;
			if (indiceIni == indice) break;
    	}
		return null;
	}
    
	/**
	 * Iterador de la tabla
   	 * @return Iterator -- que mantiene los elementos de la tabla
   	 * @see java.util.Iterator
   	 */
	public java.util.Iterator iterador() {
		return new MiIterador();
	}
	
	/**
	 * Clase privada que implementa el iterador.
	 */
	private class MiIterador implements java.util.Iterator {
		private int indice = -1;
		
		public boolean hasNext () {
			while (++indice < elementos.length) {
				if (elementos[indice] != null && !elementos[indice].equals(borrado))
					return true;
			}
			return false;
		}
		
		public Object next () {
			return elementos[indice];
		}
		
		// Metodo no soportado
		public void remove() {
			throw new UnsupportedOperationException();
		}
	}
	
	
	/**
     * Metodo para determinar el numero primo mayor al entero dado.
     * @param n -- numero positivo inicial
     * @return int -- el numero primo mayor o igual a n.
     */
    private static int sgtePrimo(int n) {
	if(n % 2 == 0)
	    n++;

	for(; !esPrimo(n); n += 2)
	    ;

	return n;
    }

    /**
     * Metodo (ineficiente) para determinar si un numero es primo.
     * @param n -- el numero a probar
     * @return boolean -- true si el numero es primo y false en otro caso.
     */
    private static boolean esPrimo(int n)  {
	if(n == 2 || n == 3)
	    return true;

	if(n == 1 || n % 2 == 0)
	    return false;

	for(int i = 3; i * i <= n; i += 2)
	    if(n % i == 0)
		return false;

	return true;
    }
}