import java.util.Iterator;
import java.util.Scanner;
import java.io.*;
/**
 * Clase para crear y manipular una lista de peliculas ordenadas por calificacion.
 * @author Yazmin Gonzalez Bayardo
 * @version Marzo 2015
 */
public class ListaPeliculas {
	
	private ListaOrdenada listaPeliculas;
	private final String nomArchivoSer;
	
	private static Scanner in = new Scanner(System.in);
	private static Pelicula pelicula = null;
	
    /**  
     * Metodo para construir una lista de peliculas ordenada por calificacion. 
	 * Recibe el nombre del archivo donde se graba y almacena los objetos Pelicula.
	 * Verifica si existe el archivo, si es el caso se lee el registro de peliculas y se almacenan en la lista ordenada de peliculas.
	 * @param nomArchSer - nombre del archivo donde se almacenan objetos serializables Pelicula.
     */
	public ListaPeliculas(String nomArchSer) {
		
		listaPeliculas = new ListaOrdenada(new ComparaCalificacion());
		
		this.nomArchivoSer = nomArchSer;
		File archivoPeliculas = new File(nomArchivoSer);
		
		if(archivoPeliculas.exists()){
			if(!archivoPeliculas.canRead()) {
				throw new RuntimeException("No es posible leer el archivo "+nomArchivoSer);
			}
			if(!archivoPeliculas.canWrite()) {
				throw new RuntimeException("No es posible escribir en el archivo "+nomArchivoSer);
			}
			
			leerPeliculas();
		}
	}
	
	/**
	 * Metodo para deserealizar un registro de peliculas
	 */
	public void leerPeliculas() {
		ObjectInputStream objIn = null;
		try {
			objIn = new ObjectInputStream(new FileInputStream(nomArchivoSer));
			Object objeto;
			do{
				objeto = objIn.readObject();
				if(objeto != null) {
					listaPeliculas.agregar((Pelicula)objeto);
				}
			}while(objeto != null);
		} catch(java.lang.ClassNotFoundException e) {
			
		} catch(java.io.EOFException e) {
			
		} catch(IOException e) {
			System.out.println("Error al leer el archivo: "+e);
		} finally {
			if(objIn != null) {
				try {
					objIn.close();
				}catch( IOException e) {
					
				}
			}
		}
	}
	
	/**
	 * Metodo para serealizar los objetos Pelicula almacenados en la lista ordenada de peliculas
	 */
	public void guardarPeliculas() {
		
		ObjectOutputStream objOut = null;
		try {
			objOut = new ObjectOutputStream(new FileOutputStream(nomArchivoSer));
			Pelicula pelicula;
			for(Iterator it = listaPeliculas.iterador(); it.hasNext();) {
				pelicula = (Pelicula)it.next();
				objOut.writeObject(pelicula);
			} 
		} catch(NotSerializableException e) {
			System.out.println("Error al grabar en el archivo. Objeto no serializable: "+e);
		} catch(IOException e) {
			System.out.println("Error al grabar en el archivo: "+e);
		} finally {
			if(objOut != null) { 
				try{ objOut.close(); } catch(IOException e){}
			}
		}
	}
	
    /**  
     * Metodo para agregar una pelicula a la lista ordenada
	 * @param titulo - nombre de la pelicula que se desea agregar
	 * @param calif - calificacion de la pelicula
     */
	public void agregarPelicula(String titulo, int calif) {
		pelicula = new Pelicula(titulo,calif);
		listaPeliculas.eliminar(pelicula); //En caso de que ya exista, se elimina para agregarla con la mas reciente calificacion
		listaPeliculas.agregar(pelicula);
		System.out.println("\nLa pelicula ha sido agregada.");
	}
	
    /**  
     * Metodo para agregar a la lista ordenada de peliculas los datos de las peliculas contenidas en un archivo de texto
	 * @param nombreArchivo - archivo de texto que contiene datos de peliculas
     */
	public void agregarArchivoPeliculas(String nombreArchivo) {
		
		try {

			Scanner sc = new Scanner(new File(nombreArchivo));
			
			while(sc.hasNextLine()) {
				String renglon = sc.nextLine();
				String[] datosPelicula = renglon.split(",");
				if(datosPelicula.length != 2)
					throw new RuntimeException("Registro de pelicula con datos incompletos.");
				
				String titulo = datosPelicula[0];
				int calificacion = Integer.parseInt(datosPelicula[1]);			
				if(calificacion < 1 || calificacion > 10)
					throw new RuntimeException("Registro de pelicula con calificacion invalida.");
				
				pelicula = new Pelicula(titulo,calificacion);
				listaPeliculas.eliminar(pelicula); //En caso de que ya exista, se elimina para agregarla con la mas reciente calificacion
				listaPeliculas.agregar(pelicula);
			}
			System.out.println("\nSe cargo el archivo de peliculas.");
		} catch(FileNotFoundException e) {
			System.out.println("Archivo no encontrado.");
		}catch(Exception e){
			System.out.println(e);
		}
	}
	
    /**  
     * Metodo para imprimir en pantalle el contenido de la lista de peliculas
     */
	public void imprimirLista() {
		int i = 1;
		System.out.println("\n   	 PELICULAS");
		System.out.println("   Titulo          Calificacion");
		for(Iterator it = listaPeliculas.iterador(); it.hasNext();) {
			pelicula = (Pelicula)it.next();
			System.out.println(i + ". " + pelicula);
			i++;
		}
	}
	
    /**  
     * Metodo para consultar la informacion de una pelicula
	 * @param titulo - nombre de la pelicula que se desea consultar
     */
	public void consultarPelicula(String titulo) {
		Nodo resul = listaPeliculas.buscar(new Pelicula(titulo));
		if(resul != null){
			System.out.println("\n   Titulo          Calificacion");
			System.out.println("   "+(Pelicula)resul.elemento);
		}else
			System.out.println("\nPelicula no encontrada");
	}
	
    /**  
     * Metodo para modificar la calificacion de una pelicula
	 * @param titulo - nombre de la pelicula cuya calificacion se desea modificar
	 * @param calif - nueva calificacion de la pelicula
     */
	public void modificarCalificacion(String titulo, int calif) {
		Pelicula peli = new Pelicula(titulo,calif);
		Nodo resul = listaPeliculas.buscar(peli);
		
		if(resul != null) {
			listaPeliculas.eliminar(peli);
			listaPeliculas.agregar(peli);
			System.out.println("\nCalificacion actualizada");
		}else
			System.out.println("\nPelicula no encontrada");
	}
	
    /**  
     * Metodo para eliminar una pelicula a la lista ordenada
	 * @param titulo - nombre de la pelicula que se desea eliminar
     */
	public void eliminarPelicula(String titulo) {
		pelicula = new Pelicula(titulo);
		if(listaPeliculas.contiene(pelicula)){
			listaPeliculas.eliminar(pelicula);
			System.out.println("\nPelicula eliminada");
		}else
			System.out.println("\nPelicula no encontrada, no es posible eliminarla");
	}
	
    /**  
     * Metodo para agregar una pelicula a la lista ordenada
	 * @param confirmacion - cadena con respuesta de confirmacion del usuario para borrar o no borrar la lista
     */
	public void borrarLista(String confirmacion) {
		if(confirmacion.equals("si")){
			listaPeliculas.vaciar();
			System.out.println("\nLista borrada");
		}else{
			if(confirmacion.equals("no"))
				System.out.println("\nOk");
			else
				System.out.println("\nRespuesta invalida");
		}
	}
	
	/**
	 * Menu de opciones para trabajar con una lista ordenada de peliculas
	 */
	public void menu() {
		System.out.println("\nLISTA DE PELICULAS");
		System.out.println("1. Agregar una pelicula");
		System.out.println("2. Agregar peliculas desde un archivo");
		System.out.println("3. Mostrar todas las peliculas");
		System.out.println("4. Consultar una pelicula");
		System.out.println("5. Actualizar calificacion");
		System.out.println("6. Eliminar pelicula");
		System.out.println("7. Borrar lista");
		System.out.println("0. Terminar");
		System.out.println("\nElige una opcion");
	}
	
	/**
	 * Metodo para verificar el valor de un calificacion y solicitar el dato hasta que sea un valor valido 
	 * @param calif - calificacion por verificar
	 * @param mensaje - mensaje que se muestra al usuario
	 * @return int - calificacion validada
	 */ 
	public int validaCalificacion(int calif, String mensaje) {
		int calificacion = calif;
		while(calificacion<1 || calificacion >10){
			System.out.println(mensaje);
		 	calificacion = Integer.parseInt(in.nextLine());
		}
		return calificacion;
	}
	
	/**
	 * Metodo para ejecutar las acciones correspondientes a la opcion seleccionada
	 * @param opcion - opcion elegida por el usuario
	 */
	public void realizarSeleccion(int opcion) {
		String titulo;
		int i, calificacion;
		Nodo resul;
		
		switch(opcion)
		{
			case 0:
				break;
				
			case 1:
				System.out.println("\nIntroduce el titulo de la pelicula");
				titulo = in.nextLine();
				System.out.println("Introduce la calificacion de la pelicula");
				calificacion = Integer.parseInt(in.nextLine());
				
				if(calificacion < 1 || calificacion > 10)
					calificacion = validaCalificacion(calificacion,"El valor de la calificacion debe ser un entero del rango 1 a 10. \nIntroduce la calificacion de la pelicula");
		
				agregarPelicula(titulo,calificacion);
				break;
			
			case 2:
				System.out.println("\nIntroduce el nombre del archivo");
				String nomArch = in.nextLine();
				agregarArchivoPeliculas(nomArch);
				break;
				
			case 3:
				imprimirLista();
				break;
			
			case 4:
				System.out.println("Introduce el titulo de la pelicula que quieres consultar");
				titulo = in.nextLine();
				consultarPelicula(titulo);
				break;
			
			case 5:
				System.out.println("Introduce el titulo de la pelicula por actualizar");
				titulo = in.nextLine();
				System.out.println("Introduce la nueva calificacion de la pelicula");
				calificacion = Integer.parseInt(in.nextLine());
				
				if(calificacion < 1 || calificacion > 10)
					calificacion = validaCalificacion(calificacion,"El valor de la calificacion debe ser un entero del rango 1 a 10. \nIntroduce la calificacion de la pelicula");

				modificarCalificacion(titulo,calificacion);
				break;
			
			case 6:
				System.out.println("Introduce el titulo de la pelicula por eliminar");
				titulo = in.nextLine();
				eliminarPelicula(titulo);
				break;
				
			case 7:
				System.out.println("¿Estas seguro de borrar la lista?... si o no");
				String respuesta = (in.nextLine()).toLowerCase().trim();
				borrarLista(respuesta);
				break;
					
			default:
				System.out.println("Opcion invalida\n");
		}
	}

	public static void main(String[] args) {
		
		int opcion;
		String nombreArchivo = (args.length != 0)? args[0] : "misPeliculas.ser";
		
		ListaPeliculas topPeliculas = new ListaPeliculas(nombreArchivo);
		
		try{
			do{
				topPeliculas.menu();
				opcion = Integer.parseInt(in.nextLine());
				topPeliculas.realizarSeleccion(opcion);
			}while(opcion!=0);
			
			topPeliculas.guardarPeliculas();
			System.out.println("¡Hasta luego!");
		} catch(Exception e) {
			System.out.println(e);
		}
	}
}